// Include class declaration
#include "Scripted_Wallpaper.h"

// Include additional requirements
#include <QFile>			// Checking script suplied wallpaper for existence
#include <QUrl>				// Load wallpaper/script path from script/settings
#include <QPainter>			// Painting/Moving the wallpaper

#include <iostream>			// For simple debugging


// Debug macro (one line code only!)
#ifdef DEBUG
	#define D if(1) 
#else
	#define D if(0)  
#endif


/***********************************************************************/
Scripted_Wallpaper::Scripted_Wallpaper(QObject *parent, const QVariantList &args)
		// Init parent
		: Plasma::Wallpaper(parent, args)
	{
	// Debug spam
	D std::cout << "[Start] Scripted_Wallpaper..." << std::endl;
		
	// Setup script callback (success / failure)
	connect(&(this->scriptProcess), SIGNAL(finished(int, QProcess::ExitStatus)),	this, 	SLOT(scriptFinished()));
	connect(&(this->scriptProcess), SIGNAL(error(QProcess::ProcessError)), 			this, 	SLOT(scriptError()));
	
	// Wallpaper has been rendered, attach it
	connect(this, 					SIGNAL(renderCompleted(QImage)), 				this, 	SLOT(renderCompleted(QImage)));
	
	// Setup timer (calls script every delay seconds)
	this->scriptTimer.setSingleShot(true);
	connect(&(this->scriptTimer), 	SIGNAL(timeout()), 								this, 	SLOT(queryScript()));
	
	// Debug spam
	D std::cout << "[End] Scripted_Wallpaper..." << std::endl;
}
Scripted_Wallpaper::~Scripted_Wallpaper() {
	// Debug spam
	D std::cout << "[Start] ~Scripted_Wallpaper..." << std::endl;
	
	// Stop script processing
    if (this->scriptProcess.state() == QProcess::Starting || this->scriptProcess.state() == QProcess::Running) {
        this->scriptProcess.close();
    }
    
    // Debug spam
	D std::cout << "[End] ~Scripted_Wallpaper..." << std::endl;
}
/***********************************************************************/


/***********************************************************************/ 
void Scripted_Wallpaper::save(KConfigGroup &config) {
	// Debug spam
	D std::cout << "[Start] save..." << std::endl;
	
	// Save to config
	config.writeEntry("fallbackResize", 	(int) this->fallbackResize);
	config.writeEntry("scriptPath", 		this->scriptPath);
	config.writeEntry("scriptArguments", 	this->scriptArguments);
	config.writeEntry("updateDelay",		this->updateDelay);
	config.writeEntry("fallbackPath", 		this->fallbackPath);
	config.writeEntry("fallbackColor", 		this->fallbackColor);
	
	// Debug spam
	D std::cout << "[End] save..." << std::endl;
}
/***********************************************************************/


/***********************************************************************/
void Scripted_Wallpaper::paint(QPainter *painter, const QRectF& exposedRect) {
	// Debug spam
	D std::cout << "[Start] paint..." << std::endl;
	
	// Dimension changed?
	if (this->previousSize != this->boundingRect().size()) {
		// Remember new size
		this->previousSize = this->boundingRect().size();
		
		// Update wallpaper
		this->renderWallpaper();

		// Draw temp desktop
		painter->fillRect(exposedRect, this->color);
		
		// Debug spam
	D std::cout << "[End Size] paint..." << std::endl;
	}
	
	// Update desktop
	else {
		// Display wallpaper
		painter->drawPixmap(exposedRect, this->wallpaperPixmap, exposedRect.translated(-this->boundingRect().topLeft()));
		
		// Debug spam
	D std::cout << "[End Draw] paint..." << std::endl;
	}
}
/***********************************************************************/


/***********************************************************************/
QWidget* Scripted_Wallpaper::createConfigurationInterface(QWidget* parent) {
	// Debug spam
	D std::cout << "[Start] createConfigurationInterface..." << std::endl;
	
	// Create new widget and setup config layout
	QWidget *widget = new QWidget(parent);
	this->configWidget.setupUi(widget);
	
	// Helper variable
	QTime 	delay = QTime().addSecs(this->updateDelay);
	int 	mode;
	
	// ... and fill positioning with values
	this->configWidget.fallbackResize->addItem(i18n("Scaled & Cropped"), 			ScaledAndCroppedResize);
	this->configWidget.fallbackResize->addItem(i18n("Scaled"), 						ScaledResize);
	this->configWidget.fallbackResize->addItem(i18n("Scaled, keep proportions"), 	MaxpectResize);
	this->configWidget.fallbackResize->addItem(i18n("Centered"), 					CenteredResize);
	this->configWidget.fallbackResize->addItem(i18n("Tiled"), 						TiledResize);
	this->configWidget.fallbackResize->addItem(i18n("Center Tiled"), 				CenterTiledResize);
	
	// ... set currently selected item
	for (int id = 0; id < this->configWidget.fallbackResize->count(); ++id) {
		// Fetch entry
		mode = this->configWidget.fallbackResize->itemData(id).value<int>();
		
		// Current setting equals this setting
		if (this->fallbackResize == mode) {
			// Mark it
			this->configWidget.fallbackResize->setCurrentIndex(id);
			
			// Be done
			break;
		}
	}

	// Update GUI from internal values
	this->configWidget.scriptPath->setText(this->scriptPath); 
	this->configWidget.scriptArguments->setText(this->scriptArguments);
	this->configWidget.updateDelay->setTime(delay);
	this->configWidget.fallbackPath->setText(this->fallbackPath);
	this->configWidget.fallbackColor->setColor(this->fallbackColor);

	// Connect GUI update to internal callback
	connect(this->configWidget.fallbackResize, 	SIGNAL(currentIndexChanged(int)),	this,	SLOT(settingsModified()));
	connect(this->configWidget.scriptPath, 		SIGNAL(textChanged(QString)), 		this, 	SLOT(settingsModified()));
	connect(this->configWidget.scriptArguments,	SIGNAL(textChanged(QString)), 		this, 	SLOT(settingsModified()));
	connect(this->configWidget.updateDelay, 	SIGNAL(timeChanged(QTime)),			this, 	SLOT(settingsModified()));
	connect(this->configWidget.fallbackPath, 	SIGNAL(textChanged(QString)),		this, 	SLOT(settingsModified()));
	connect(this->configWidget.fallbackColor, 	SIGNAL(changed(QColor)),			this, 	SLOT(settingsModified()));

	// Connect internal changes to GUI update
	connect(this, 								SIGNAL(settingsChanged(bool)), 		parent,	SLOT(settingsChanged(bool)));
	
	// Debug spam
	D std::cout << "[End] createConfigurationInterface..." << std::endl;

	// Return config widget
	return widget;
}
/***********************************************************************/


/***********************************************************************/
void Scripted_Wallpaper::init(const KConfigGroup &config) {
	// Debug spam
	D std::cout << "[Start] Init..." << std::endl;
	
	// Reset timer
	if(this->scriptTimer.isActive())
		this->scriptTimer.stop();

	// Load config values (or use defaults)
	this->fallbackResize 	= (ResizeMethod) config.readEntry("fallbackResize", 	(int) ScaledResize);
	this->scriptPath 		=				 config.readEntry("scriptPath", 		QString());
	this->scriptArguments 	= 				 config.readEntry("scriptArguments", 	QString());
	this->updateDelay 		= 				 config.readEntry("updateDelay", 		30 * 60);
	this->fallbackPath 		= 				 config.readEntry("fallbackPath", 		QString());
	this->fallbackColor	 	= 				 config.readEntry("fallbackColor", 		QColor(0.0, 0.0, 0.0));
	
	// Initialize
	this->wallpaperPath		= this->fallbackPath;
	this->resize			= this->fallbackResize;
	this->color				= this->fallbackColor;
	
	// Update setting
	this->setResizeMethodHint(this->resize);
	this->previousSize 		= this->boundingRect().size();
	
	// Initialize
    this->setFallbackWallpaper();
    
    // Query wallpaper update (will be handled by scriptFinished)
    this->setupScript();
    this->queryScript();

	// Repaint wallpaper
	emit update(this->boundingRect());
	
	// Debug spam
	D std::cout << "[End] Init..." << std::endl;
}
/***********************************************************************/


/***********************************************************************/
void Scripted_Wallpaper::settingsModified() {
	// Debug spam
	D std::cout << "[Start] settingsModified..." << std::endl;
	
	// Not modified by default
	bool modified = false;

	// fallbackResize changed?
	int id 		= this->configWidget.fallbackResize->currentIndex();
	int mode 	= this->configWidget.fallbackResize->itemData(id).value<int>();
	if (this->fallbackResize != (ResizeMethod) mode) {
		// Remember update
		this->fallbackResize = (ResizeMethod) mode;

		// Reload wallpaper (asyncron via setWallpaper(QImage))
		this->renderWallpaper();
		
		// Reload wallpaper (asyncron via setWallpaper(QImage))
		this->renderWallpaper();

		// Modified!
		modified = true;
	}

	// scriptPath changed?
	QString script = this->configWidget.scriptPath->text(); 
	if (this->scriptPath != script) {
		// Remember update
		this->scriptPath = script;

		// Update and run script (asyncron via scriptFinished(...) or scriptError(...))
		this->setupScript();
		this->queryScript();

		// Modified!
		modified = true;
	}

	// scriptArguments changed?
	QString arguments = this->configWidget.scriptArguments->text();
	if (this->scriptArguments != arguments) {
		// Remember update
		this->scriptArguments = arguments;

		// Update and run script (asyncron via scriptFinished(...) or scriptError(...))
		this->setupScript();
		this->queryScript();

		// Modified!
		modified = true;
	}

	// updateDelay changed?
	unsigned int seconds = qAbs(this->configWidget.updateDelay->time().secsTo(QTime()));
	if (this->updateDelay != seconds) {
		// Remember update
		this->updateDelay = seconds;
		
		// No update less than once per second!
		if(this->updateDelay == 0)
			this->updateDelay = 1;
			
		// Run script (asyncron via scriptFinished(...) or scriptError(...))
		this->queryScript();
		
		// Modified!
		modified = true;
	}

	// fallbackPath changed?
	if (this->fallbackPath != this->configWidget.fallbackPath->text()) { 
		// Remember update
		this->fallbackPath = this->configWidget.fallbackPath->text(); 
		
		// Reload wallpaper (asyncron via setWallpaper(QImage))
		this->renderWallpaper();
		
		// Modified!
		modified = true;
	}
	
	// fallbackColor changed?
	if (this->fallbackColor != this->configWidget.fallbackColor->color()) { 
		// Remember update
		this->fallbackColor = this->configWidget.fallbackColor->color(); 
		
		// Reload wallpaper (asyncron via setWallpaper(QImage))
		this->renderWallpaper();
		
		// Modified!
		modified = true;
	}

	// Repaint wallpaper and also inform KDE
	if(modified) {
		emit settingsChanged(true);
		emit update(this->boundingRect());
	}
	
	// Debug spam
	D std::cout << "[End] settingsModified..." << std::endl;
}
/***********************************************************************/


/***********************************************************************/
void Scripted_Wallpaper::renderCompleted(const QImage &image) {
	// Debug spam
	D std::cout << "[Start] renderCompleted..." << std::endl;
	
	// Update wallpaper
	this->wallpaperPixmap = QPixmap::fromImage(image);
	
	// Repaint wallpaper
	emit update(this->boundingRect());
	
	// Debug spam
	D std::cout << "[End] renderCompleted..." << std::endl;
}
/***********************************************************************/


/***********************************************************************/
void Scripted_Wallpaper::scriptFinished() {
	// Debug spam
	D std::cout << "[Start] scriptFinished..." << std::endl;
	
	// Use fallback first
	this->wallpaperPath	= this->fallbackPath;
	this->resize		= this->fallbackResize;
	this->color			= this->fallbackColor;
	
	// Read script output and convert unicode
	QByteArray output 	= this->scriptProcess.readAll();
	
	// 
	if (output.size() > 1) {
		// Trimm spaces
		QString arguments	= QString::fromLocal8Bit(output).trimmed();
		
		// Debug spam
		D std::cout << "Output: " << arguments.toUtf8().constData() << std::endl;
		
		// Fetch arguments
		if (arguments.length() > 0) {
			QStringList list 		= arguments.split("\n", QString::SkipEmptyParts);
			
			// Parse arguments #1 - Wallaper
			if(list.size() > 0 && !list.at(0).isEmpty()) {
				// Absolute path?
				if (QFile::exists(list.at(0))) 
					this->wallpaperPath	= list.at(0);
				
				// Relative path!
				else {
					QUrl relativeURL 	= QUrl().fromLocalFile(list.at(0));
					QUrl scriptURL 		= QUrl().fromLocalFile(this->scriptPath);
					this->wallpaperPath	= scriptURL.resolved(relativeURL).toLocalFile();
				}
			}
			
			// Parse arguments #2 - Resize mode
			if (list.size() > 1 && !list.at(1).isEmpty())
				this->resize		= this->resizeMethodStringToEnum(list.at(1));
				
			// Parse arguments #3 - BG color
			if (list.size() > 2 && !list.at(2).isEmpty()) {
				// Fetch colors
				arguments 			= list.at(2);
				list 				= arguments.split(" ", QString::SkipEmptyParts);
				
				// 3 Color components?
				if (list.size() == 3)
					this->color		= QColor(list.at(0).toInt(), list.at(1).toInt(), list.at(2).toInt());
			}
		}
	}
	
	// Debug spam
	D std::cout << "Wallpaper Path: " << this->wallpaperPath.toUtf8().constData() << std::endl;
	D std::cout << "Resize Mode: " << this->resizeMethodEnumToString(this->resize).toUtf8().constData() << std::endl;
	D std::cout << "BG Color: " << this->color.red() << "R, " << this->color.green() << "G, " << this->color.blue() << "B" << std::endl;

	// Refresh new wallpaper
	this->renderWallpaper();
	
	// Debug spam
	D std::cout << "[End] scriptFinished..." << std::endl;
}
void Scripted_Wallpaper::scriptError() {
	// Debug spam
	D std::cout << "[Start] scriptError..." << std::endl;
	
	// Delegate
	this->setFallbackWallpaper();
	
	// Debug spam
	D std::cout << "[End] scriptError..." << std::endl;
}
/***********************************************************************/


/***********************************************************************/
void Scripted_Wallpaper::queryScript() {
	// Debug spam
	D std::cout << "[Start] queryScript..." << std::endl;
	
	// Restart script 
	if (this->scriptProcess.state() == QProcess::NotRunning) {
		this->scriptProcess.close();
		this->scriptProcess.start();
	}
	
	// Update timer
	int delay = qMax((int) this->updateDelay * 1000, 1000);
	this->scriptTimer.start(delay);
	
	// Debug spam
	D std::cout << "[End] queryScript..." << std::endl;
}
/***********************************************************************/


/***********************************************************************/
void Scripted_Wallpaper::setupScript() {
	// Debug spam
	D std::cout << "[Start] setupScript..." << std::endl;
	
	// Stop old script
	if (this->scriptProcess.state() == QProcess::Running) 
		this->scriptProcess.close();

	// Set script and parameters
	this->scriptProcess.setProgram(this->scriptPath, this->getArgumentsList());
	
	// Set working directory
	QString scriptDirectory = QUrl().fromLocalFile(this->scriptPath).resolved(QUrl(".")).toLocalFile();
	this->scriptProcess.setWorkingDirectory(scriptDirectory);
	
	// Set output method
	this->scriptProcess.setOutputChannelMode(KProcess::OnlyStdoutChannel);
	
	// Debug spam
	D std::cout << "[End] setupScript..." << std::endl;
}
/***********************************************************************/


/***********************************************************************/
void Scripted_Wallpaper::renderWallpaper() {
	// Debug spam
	D std::cout << "[Start] renderWallpaper..." << std::endl;
	
	// Display script wallpaper?
	if (QFile::exists(this->wallpaperPath)) 
		this->render(this->wallpaperPath, this->boundingRect().size().toSize(), this->resize, this->color);
		
	// ... or Fallback?
	else 
		this->setFallbackWallpaper();
		
	// Debug spam
	D std::cout << "[End] renderWallpaper..." << std::endl;
}
/***********************************************************************/


/***********************************************************************/
void Scripted_Wallpaper::setFallbackWallpaper() {
	// Debug spam
	D std::cout << "[Start] setFallbackWallpaper..." << std::endl;
	
	if (QFile::exists(this->fallbackPath)) 
		this->render(this->fallbackPath, this->boundingRect().size().toSize(), this->resize, this->color);
	
	else {
		// Reset wallpaper to fallback color
		this->wallpaperPixmap = QPixmap(this->boundingRect().size().toSize());
		this->wallpaperPixmap.fill(this->fallbackColor);
		
		// Repaint wallpaper
		emit update(this->boundingRect());
	}
	
	// Debug spam
	D std::cout << "[End] setFallbackWallpaper..." << std::endl;
}
/***********************************************************************/


/***********************************************************************/
QStringList Scripted_Wallpaper::getArgumentsList() {
	// Debug spam
	D std::cout << "[Start] getArgumentsList..." << std::endl;
	
	// Temp variable for parsing
	QString parsedArguments = this->scriptArguments;
	parsedArguments.replace("%R", this->resizeMethodEnumToString(this->resize));

	// Will be returned
	QStringList argumentsList;

	// Loop variables
	int charPosition = 0;
	int argumentStart = 0;
	bool insideQuotes = false;

	// Loop over string
	while(charPosition < parsedArguments.length()) {
		// Query character
		QChar c = parsedArguments.at(charPosition);
		
		// Test for quotes
		if (c == '"') {
			insideQuotes = !insideQuotes;
		// Next parameter	
		} else if (c == ' ') {
			if(!insideQuotes) {
				QString argument = parsedArguments.mid(argumentStart, charPosition - argumentStart);
				argument.replace('"', "");
				argumentsList.append(argument);
				argumentStart = charPosition + 1;
			}
		}
		
		// Continue
		charPosition++;
	}

	// Extra case for last argument
	QString argument = parsedArguments.mid(argumentStart, charPosition - argumentStart);
	argument.replace('"', "");
	argumentsList.append(argument);
	
	// Debug spam
	D std::cout << "[End] getArgumentsList..." << std::endl;

	// Return list
	return argumentsList;
}
/***********************************************************************/


/***********************************************************************/
QString Scripted_Wallpaper::resizeMethodEnumToString(ResizeMethod input) {
	// Might be replaced by a std::map
	switch (input) {
		case ScaledResize:
			return "ScaledResize";
		case CenteredResize:
			return "CenteredResize";
		case ScaledAndCroppedResize:
			return "ScaledAndCroppedResize";
		case TiledResize:
			return "TiledResize";
		case CenterTiledResize:
			return "CenterTiledResize";
		case MaxpectResize:
			return "MaxpectResize";
	}
	
	// Fallback
	return this->resizeMethodEnumToString(this->fallbackResize);
}
Scripted_Wallpaper::ResizeMethod Scripted_Wallpaper::resizeMethodStringToEnum(QString input) {
	// Might be replaced by a std::map
	if 		(input == "ScaledResize")
			return ScaledResize;
	else if (input == "CenteredResize")
			return CenteredResize;
	else if (input == "ScaledAndCroppedResize")
			return ScaledAndCroppedResize;
	else if (input == "TiledResize")
			return TiledResize;
	else if (input == "CenterTiledResize")
			return CenterTiledResize;
	else if (input == "MaxpectResize")
			return MaxpectResize;
	
	// Fallback
	return this->fallbackResize;
}
/***********************************************************************/


// Include Qt converted Scripted_Wallpaper.h
#include "Scripted_Wallpaper.moc"
