// Header protection
#ifndef PLASMA_WALLPAPER_SCRIPTED_WALLPAPER
#define PLASMA_WALLPAPER_SCRIPTED_WALLPAPER


// Include Required headers
#include <Plasma/Wallpaper>		// Required for any wallpaper script
#include <QTimer>				// Keep track of last activation

#include <KProcess>				// Run a timer in the background

#include "ui_config.h" 			// Config generated by QT from config.ui


// Current wallpaper state
enum WPState { WPScript, WPFallback, WPBlank };
 

// Derive own wallpaper script implementation from Plasma::Wallpaper
class Scripted_Wallpaper : public Plasma::Wallpaper {

// Qt macro (important to allow qt callbacks to work)
Q_OBJECT

    
// Important public members
public:
	/*
	 * [Constructor] Scripted_Wallpaper(parent, args)
	 *  Initialize timer, internal variables nd callbacks.
	 * Arguments:
	 *  parent - QObject parent; you probably want to pass in 0
	 *  args - a list of strings containing one entry: the service id
	 */
	Scripted_Wallpaper(QObject* parent, const QVariantList& args);
	~Scripted_Wallpaper();
	
	/*
	 * save(config)
	 *  This method is called when settings need to be saved.  
	 * Arguments:
	 *  config - The Config group to save settings to
	 */
	virtual void save(KConfigGroup &config);
	
	/*
	 * paint(painter, exposedRect)
	 *  This method is called when the wallpaper should be re-painted. (eg.by update(...))
	 * Arguments:
	 *   painter - The QPainter to use to do the painting 
	 *   exposedRect - 	The rectangle to paint within 
	 */
	void paint(QPainter* painter, const QRectF& exposedRect);
	
	/*
	 * createConfigurationInterface(parent)
	 *  Returns a widget that can be used to configure the options (if any) associated with this wallpaper.
	 * Arguments:
	 *  parent - Parent Qt widget to which our config will be attached to
	 * Returns:
	 *  Qt widget to use for configuring
	 */
	virtual QWidget* createConfigurationInterface(QWidget* parent); 
	
	
// Qt macro ([OUT] qt callbacks)
Q_SIGNALS:
	/*
	 * settingsChanged(modified)
	 *  Signal parent that a config setting was changed.
	 * Arguments:
	 *  modified - Should be set to true if new GUI settings differe from old
	 */
	void settingsChanged(bool modified); 


// Important deriveable members
protected:
	/*
	 * init(config)
	 *  This method is called once the wallpaper is loaded or mode is changed. 
	 * Arguments:
	 *  config - The Config group to load settings from
	 */
	virtual void init(const KConfigGroup &config);


// Qt macro ([IN] qt callbacks)
protected slots: 
	/*
	 * settingsModified()
	 *  [IN] ]Callback when the settings where modified by the GUI and the wallpaper need an update.
	 */
	void settingsModified(); 
	
	
// Qt macro ([IN] qt callbacks)
private slots:
	/*
	 * renderCompleted(image)
	 *  Set wallpaper from QImage.
	 * Asyncron: 
	 *  After a render() call succeeded.
	 */
	void renderCompleted(const QImage &image);

	/*
	 * scriptFinished()
	 *  Get and parse script output and load wallpaper.
	 * Asyncron: 
	 *  After a runScript() called and script succeeded.
	 */
	void scriptFinished();
	
	/*
	 * scriptError()
	 *  Use fallback wallpaper.
	 * Asyncron: 
	 *  After a runScript() called and script failed.
	 */
	void scriptError();
	
	/*
	 * runScript()
	 *  Run the script. See script*().
	 * Asyncron:
	 *  After each scriptTimer timeOut()
	 */
	void queryScript(); 
	
	
// Internal methods
private:	
	/*
	 * setupScript()
	 *  Setup script to be run later.
	 */
	void setupScript();
	
	/*
	 * redrawWallpaper()
	 *  Refresh wallpaper image (no script update!)
	 */
	void renderWallpaper();
	
	/*
	 * setFallbackWallpaper()
	 *  Set wallpaper to fallback image or blank
	 */
	void setFallbackWallpaper();
	
	/*
	 * getParsedArgumentsList()
	 *  Parse GUI arguments
	 */
	QStringList getArgumentsList();
	
	/*
	 * resizeMethodEnumToString(ResizeMethod)
	 * resizeMethodStringToEnum(QString)
	 *  Convert from/to ResizeMethod/QString.
	 * Arguments:
	 *  ResizeMethod - Input to convert
	 *  QString - Input to convert
	 * Returns
	 *  QString - Converted
	 *  ResizeMethod - Converted
	 */
	QString resizeMethodEnumToString(ResizeMethod input);
	ResizeMethod resizeMethodStringToEnum(QString input);
	
	
// Internal variables
private:
	Ui::Config 		configWidget; 			// Config-GUI definition

	ResizeMethod 	fallbackResize;			// Fallback alignment/zoom method
	QString 		scriptPath;				// Location of script
	QString 		scriptArguments;		// Arguments to pass to script
	unsigned int	updateDelay;			// Delay between script/wallpaper update (seconds)
	QString 		fallbackPath;			// Fallback wallpaper location
	QColor			fallbackColor;			// Fallback background fill color

	KProcess 		scriptProcess;			// Script call process
	QTimer 			scriptTimer;			// Keep track of last update

	WPState			state;					// Script wallpaper, fallback image or blank
	QPixmap 		wallpaperPixmap;		// Current wallpaper memory
	QString 		wallpaperPath;			// Remember current WP location
	QColor			color;					// Script bg color
	ResizeMethod	resize;					// Script resize mode

	QSizeF 			previousSize;			// Keep track of eg. resolution change
};


// Tell KDE about plasma wallpaper plugin
K_EXPORT_PLASMA_WALLPAPER(scripted_wallpaper, Scripted_Wallpaper)
 

// Header protection
#endif
